from temporalio.worker.workflow_sandbox import SandboxRestrictions

SANDBOX_PASSTHROUGH_MODULES = ("mistralai.workflows",)

SANDBOX_RESTRICTIONS = SandboxRestrictions.default.with_passthrough_modules(*SANDBOX_PASSTHROUGH_MODULES)
